#
# Copyright (c) 2002-2004 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Handles creation of drawing entities (circles, lines, chamfers, etc)
#

import objc
import string
from math import hypot, pi, atan2

import Globals
import ImageDocument
import Generic.tools
import Generic.util

from Foundation import *
from AppKit import NSShiftKeyMask

#
# Helper functions
#
def make_tuple(text):
    
    try:
        _tpl = eval(text)
    except:
        raise StandardError, "Invalid point: " + text
    if not isinstance(_tpl, tuple):
        raise TypeError, "Invalid point: " + str(_tpl)
    if len(_tpl) != 2:
        raise ValueError, "Invalid point: " + str(_tpl)
    return _tpl

def make_radius(text):
    try:
        _textrad = eval(text)
    except:
        raise ValueError, "Invalid radius: " + text
    _rad = _textrad
    if not isinstance(_rad, float):
        _rad = float(_textrad)
    if _rad < 1e-10:
        raise ValueError, "Invalid radius: %g" % _rad
    return _rad

def make_angle(text):
    try:
        _textangle = eval(text)
    except:
        raise StandardError, "Invalid angle: " + text
    return Generic.util.make_c_angle(_textangle)

def create_entity(doc, tool):
    _init_func = tool.getHandler("initialize")
    _image = doc.getImage()
    _image.startAction()
    try:
        tool.create(_image)
    finally:
        _image.endAction()
    _init_func(doc, tool)

Globals.s = None
Globals.l = None
Globals.c = None
Globals.t = None
#
# Points
#
     
def point_mode_init(doc, tool):
    doc.setPrompt("Click in the drawing area or enter the point as 'x, y'")
    tool.setHandler("initialize", point_mode_init)
    tool.setHandler("left_button_press", point_left_button_press_cb)
    tool.setHandler("entry_event", point_entry_event_cb)

def point_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    _viewLoc = _da.convertPoint_fromView_(_loc, None)
    tool.setPoint(_viewLoc.x, _viewLoc.y)
    create_entity(doc, tool)

def point_entry_event_cb(doc, text, tool):
    if len(text):
        _x, _y = make_tuple(text)
        tool.setPoint(_x, _y)
        create_entity(doc, tool)

#
# Segments
#
def segment_mode_init(doc, tool):
    doc.setPrompt("Click in the drawing area or enter the first point as 'x, y'")
    tool.setHandler("initialize", segment_mode_init)
    tool.setHandler("left_button_press", segment_first_left_button_press_cb)
    tool.setHandler("entry_event", segment_first_entry_event_cb)
    _image = doc.getImage()
    Globals.s = _image.getOption('LINE_STYLE')
    Globals.l = _image.getOption('LINE_TYPE')
    Globals.c = _image.getOption('LINE_COLOR')
    Globals.t = _image.getOption('LINE_THICKNESS')

def segment_first_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    _viewLoc = _da.convertPoint_fromView_(_loc, None)
    tool.setFirstPoint(_viewLoc.x, _viewLoc.y)
    _da.setTempPoint(_viewLoc)
    tool.setHandler("left_button_press", segment_second_left_button_press_cb)
    tool.setHandler("mouse_move", segment_mouse_move_cb)
    doc.setPrompt("Click in the drawing area or enter the second point as 'x, y'")

def segment_mouse_move_cb(doc, np, tool):
    _p1 = tool.getFirstPoint()
    _p2 = (np.x, np.y)
    _seg = Generic.segment.Segment(_p1, _p2, Globals.s, Globals.l, Globals.c, Globals.t)
    _da = doc.getDA()
    _da.setTempObject(_seg)
   
def segment_second_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    _sp = _da.convertPoint_fromView_(_loc, None)
    tool.setSecondPoint(_sp.x, _sp.y)
    create_entity(doc, tool)

def segment_first_entry_event_cb(doc, text, tool):
    if len(text):
        _x, _y = make_tuple(text)
        tool.setFirstPoint(_x, _y)
        doc.getDA().setTempPoint(NSMakePoint(_x, _y))
        tool.setHandler("left_button_press", segment_second_left_button_press_cb)
        tool.setHandler("mouse_move", segment_mouse_move_cb)
        tool.setHandler("entry_event", segment_second_entry_event_cb)
        doc.setPrompt("Click in the drawing area or enter the second point as 'x, y'")

def segment_second_entry_event_cb(doc, text, tool):
    if len(text):
        _x, _y = make_tuple(text)
        tool.setSecondPoint(_x, _y)
        create_entity(doc, tool)
        
#
# rectangles
#
def rectangle_mode_init(doc, tool):
    doc.setPrompt("Click in the drawing area or enter a point as 'x, y'")
    tool.setHandler("initialize", rectangle_mode_init)
    tool.setHandler("left_button_press", rectangle_first_left_button_press_cb)
    tool.setHandler("entry_event", rectangle_first_entry_event_cb)
    _image = doc.getImage()
    Globals.s = _image.getOption('LINE_STYLE')
    Globals.l = _image.getOption('LINE_TYPE')
    Globals.c = _image.getOption('LINE_COLOR')
    Globals.t = _image.getOption('LINE_THICKNESS')

def rectangle_first_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    _viewLoc = _da.convertPoint_fromView_(_loc, None)
    tool.setFirstPoint(_viewLoc.x, _viewLoc.y)
    _da.setTempPoint(_viewLoc)
    tool.setHandler("left_button_press", rectangle_second_left_button_press_cb)
    tool.setHandler("mouse_move", rectangle_mouse_move_cb)
    tool.setHandler("entry_event", rectangle_second_entry_event_cb)
    doc.setPrompt("Click in the drawing area or enter the second point as 'x, y'")

def rectangle_mouse_move_cb(doc, np, tool):
    _p1 = tool.getFirstPoint()
    _p2 = (_p1[0],np.y)
    _p3 = (np.x, np.y)
    _p4 = (np.x,_p1[1])
    _da = doc.getDA()
    _s1 = Generic.segment.Segment(_p1, _p2, Globals.s, Globals.l, Globals.c, Globals.t)
    _s2 = Generic.segment.Segment(_p2, _p3, Globals.s, Globals.l, Globals.c, Globals.t)
    _s3 = Generic.segment.Segment(_p3, _p4, Globals.s, Globals.l, Globals.c, Globals.t)
    _s4 = Generic.segment.Segment(_p4, _p1, Globals.s, Globals.l, Globals.c, Globals.t)
    _da = doc.getDA()
    _da.setTempObject(_s1)
    _da.setTempObject(_s2, False)
    _da.setTempObject(_s3, False)
    _da.setTempObject(_s4, False)
    
def rectangle_second_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    _sp = _da.convertPoint_fromView_(_loc, None)
    tool.setSecondPoint(_sp.x, _sp.y)
    create_entity(doc, tool)

def rectangle_first_entry_event_cb(doc, text, tool):
    if len(text):
        _x, _y = make_tuple(text)
        tool.setFirstPoint(_x, _y)
        tool.setHandler("left_button_press", rectangle_second_left_button_press_cb)
        tool.setHandler("mouse_move", rectangle_mouse_move_cb)
        tool.setHandler("entry_event", rectangle_second_entry_event_cb)
        doc.setPrompt("Click in the drawing area or enter the second point as 'x, y'")
        doc.getDA().setTempPoint(NSMakePoint(_x, _y))

def rectangle_second_entry_event_cb(doc, text, tool):
    if len(text):
        _x, _y = make_tuple(text)
        tool.setSecondPoint(_x, _y)
        create_entity(doc, tool)

#
# Circles
#


# center point
def circle_center_mode_init(doc, tool):
    doc.setPrompt("Click in the drawing area or enter a point as 'x, y'")
    tool.setHandler("initialize", circle_center_mode_init)
    tool.setHandler("left_button_press", circle_center_first_left_button_press_cb)
    tool.setHandler("entry_event", circle_point_entry_event_cb)
    _image = doc.getImage()
    Globals.s = _image.getOption('LINE_STYLE')
    Globals.l = _image.getOption('LINE_TYPE')
    Globals.c = _image.getOption('LINE_COLOR')
    Globals.t = _image.getOption('LINE_THICKNESS')
    
def circle_center_first_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    _viewLoc = _da.convertPoint_fromView_(_loc, None)
    tool.setCenter(_viewLoc.x, _viewLoc.y)
    tool.setHandler("left_button_press", circle_center_second_left_button_press_cb)
    tool.setHandler("mouse_move", circle_center_mouse_move_cb)
    tool.setHandler("entry_event", circle_radius_entry_event_cb)
    doc.setPrompt("Click in the drawing area or enter the radius")
    _da.setTempPoint(_viewLoc)
    
def circle_center_second_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    _sp = _da.convertPoint_fromView_(_loc, None)
    _cx, _cy = tool.getCenter()
    _radius = hypot((_cx - _sp.x), (_cy - _sp.y))
    tool.setRadius(_radius)
    create_entity(doc, tool)
    
def circle_center_mouse_move_cb(doc, np, tool):
    _p1 = tool.getCenter()
    _r = hypot((_p1[0] - np.x), (_p1[1] - np.y))
    if _r > 0.0:
        _circle = Generic.circle.Circle(_p1, _r, Globals.s, Globals.l, Globals.c, Globals.t)
        _da = doc.getDA()
        _da.setTempObject(_circle)
    
def circle_point_entry_event_cb(doc, text, tool):
    if len(text):
        _x, _y = make_tuple(text)
        tool.setCenter(_x, _y)
        tool.setHandler("left_button_press", circle_center_second_left_button_press_cb)
        tool.setHandler("mouse_move", circle_center_mouse_move_cb)
        tool.setHandler("entry_event", circle_radius_entry_event_cb)
        doc.setPrompt("Click in the drawing area or enter the radius")
        doc.getDA().setTempPoint(NSMakePoint(_x, _y))
    
def circle_radius_entry_event_cb(doc, text, tool):
    if len(text):
        _r = make_radius(text)
        tool.setRadius(_r)
        create_entity(doc, tool)
 
# 2 point
def circle_tp_mode_init(doc, tool):
    doc.setPrompt("Click in the drawing area or enter a point as 'x, y'")
    tool.setHandler("initialize", circle_tp_mode_init)
    tool.setHandler("left_button_press", circle_tp_first_left_button_press_cb)
    tool.setHandler("entry_event", circle_tp_first_entry_event_cb)
    _image = doc.getImage()
    Globals.s = _image.getOption('LINE_STYLE')
    Globals.l = _image.getOption('LINE_TYPE')
    Globals.c = _image.getOption('LINE_COLOR')
    Globals.t = _image.getOption('LINE_THICKNESS')

def circle_tp_first_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    _viewLoc = _da.convertPoint_fromView_(_loc, None)
    _da.setTempPoint(_viewLoc)
    tool.setFirstPoint(_viewLoc.x, _viewLoc.y)
    tool.setHandler("left_button_press", circle_tp_second_left_button_press_cb)
    tool.setHandler("mouse_move", circle_tp_mouse_move_cb)
    tool.setHandler("entry_event", circle_tp_second_entry_event_cb)
    doc.setPrompt("Click in the drawing area or enter the second point as 'x, y'")
    
def circle_tp_second_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    _sp = _da.convertPoint_fromView_(_loc, None)
    tool.setSecondPoint(_sp.x, _sp.y)
    create_entity(doc, tool)

def circle_tp_mouse_move_cb(doc, np, tool):
    _x, _y = tool.getFirstPoint()
    _fp = NSMakePoint(_x, _y)
    if NSEqualPoints(_fp, np):
        return
    tool.setSecondPoint(np.x, np.y)
    _cp = tool.getCenter()
    _r = tool.getRadius()
    if _r is not None and _r > 0:
        _circle = Generic.circle.Circle(_cp, _r, Globals.s, Globals.l, Globals.c, Globals.t)
        _da = doc.getDA()
        _da.setTempObject(_circle)
    
def circle_tp_first_entry_event_cb(doc, text, tool):
    if len(text):
        _x, _y = make_tuple(text)
        tool.setFirstPoint(_x, _y)
        tool.setHandler("left_button_press", circle_tp_second_left_button_press_cb)
        tool.setHandler("mouse_move", circle_tp_mouse_move_cb)
        tool.setHandler("entry_event", circle_tp_second_entry_event_cb)
        doc.setPrompt("Click in the drawing area or enter the second point as 'x, y'")
        doc.getDA().setTempPoint(NSMakePoint(_x, _y))
        

def circle_tp_second_entry_event_cb(doc, text, tool):
    if len(text):
        _x, _y = make_tuple(text)
        tool.setSecondPoint(_x, _y)
        create_entity(doc, tool)

#
# Arcs
#

def arc_center_mode_init(doc, tool):
    doc.setPrompt("Click in the drawing area or enter a point as 'x, y'")
    tool.setHandler("initialize", arc_center_mode_init)
    tool.setHandler("left_button_press", arc_center_first_left_button_press_cb)
    tool.setHandler("entry_event", arc_center_entry_event_cb)
    _image = doc.getImage()
    Globals.s = _image.getOption('LINE_STYLE')
    Globals.l = _image.getOption('LINE_TYPE')
    Globals.c = _image.getOption('LINE_COLOR')
    Globals.t = _image.getOption('LINE_THICKNESS')
    
def arc_center_first_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    _viewLoc = _da.convertPoint_fromView_(_loc, None)
    _da.setTempPoint(_viewLoc)
    tool.setCenter(_viewLoc.x, _viewLoc.y)
    tool.setHandler("left_button_press", arc_center_second_left_button_press_cb)
    tool.setHandler("mouse_move", arc_radius_mouse_move_cb)
    tool.setHandler("entry_event", arc_radius_entry_event_cb)
    doc.setPrompt("Click in the drawing area or enter the radius")
    
def arc_center_second_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    _sp = _da.convertPoint_fromView_(_loc, None)
    _cx, _cy = tool.getCenter()
    _radius = hypot((_sp.x - _cx), (_sp.y - _cy))
    if _radius > 0.0:
        tool.setRadius(_radius)
        _angle = (180.0/pi) * atan2((_sp.y - _cy), (_sp.x - _cx))
        if _angle < 0.0:
            _angle = _angle + 360.0
        tool.setStartAngle(_angle)
        tool.delHandler("entry_event")
        tool.setHandler("left_button_press", arc_center_third_left_button_press_cb)
        tool.setHandler("mouse_move", arc_angle_mouse_move_cb)
        doc.setPrompt("Click in the drawing area to finish the arc")
    
def arc_center_third_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    _tp = _da.convertPoint_fromView_(_loc, None)
    _cx, _cy = tool.getCenter()
    _angle = (180.0/pi) * atan2((_tp.y - _cy), (_tp.x - _cx))
    if _angle < 0.0:
        _angle = _angle + 360.0
    tool.setEndAngle(_angle)
    create_entity(doc, tool)
    
def arc_start_angle_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    (_x, _y) = _da.convertPoint_fromView_(_loc, None)
    _tol = _da.pointSize().width
    _image = doc.getImage()
    _pt, _flag = _image.findPoint(_x, _y, _tol)
    if _pt is not None:
        _x, _y = _pt.getCoords()
    _cx, _cy = tool.getCenter()
    _angle = (180.0/pi) * atan2((_y - _cy), (_x - _cx))
    if _angle < 0.0:
        _angle = _angle + 360.0
    tool.setStartAngle(_angle)
    tool.setHandler("mouse_move", arc_angle_mouse_move_cb)
    tool.setHandler("left_button_press", arc_center_third_left_button_press_cb)
    tool.setHandler("entry_event", arc_end_angle_entry_event_cb)
    doc.setPrompt("Click in the drawing area or enter the end angle of the arc")
    

def arc_radius_mouse_move_cb(doc, np, tool):
    _p1 = tool.getCenter()
    _p2 = (np.x, np.y)
    _seg = Generic.segment.Segment(_p1, _p2, Globals.s, Globals.l, Globals.c, Globals.t)
    _da = doc.getDA()
    _da.setTempObject(_seg)

def arc_angle_mouse_move_cb(doc, np, tool):
    _cp = tool.getCenter()
    _cx, _cy = _cp
    _r = tool.getRadius()
    _sa = tool.getStartAngle()
    _ea = (180.0/pi) * atan2((np.y - _cy), (np.x - _cx))
    if _ea < 0.0:
        _ea = _ea + 360.0
    _arc = Generic.arc.Arc(_cp, _r, _sa, _ea, Globals.s, Globals.l, Globals.c, Globals.t)
    _da = doc.getDA()
    _da.setTempObject(_arc)
    
def arc_center_entry_event_cb(doc, text, tool):
    if len(text):
        _x, _y = make_tuple(text)
        tool.setCenter(_x, _y)
        tool.setHandler("left_button_press", arc_center_second_left_button_press_cb)
        tool.setHandler("mouse_move", arc_radius_mouse_move_cb)
        tool.setHandler("entry_event", arc_radius_entry_event_cb)
        doc.setPrompt("Click in the drawing area or enter the radius")
        doc.getDA().setTempPoint(NSMakePoint(_x, _y))

def arc_radius_entry_event_cb(doc, text, tool):
    if len(text):
        _r = make_radius(text)
        tool.setRadius(_r)
        tool.delHandler("mouse_move")
        tool.setHandler("left_button_press", arc_start_angle_left_button_press_cb)
        tool.setHandler("entry_event", arc_start_angle_entry_event_cb)
        doc.setPrompt("Click in the drawing area or enter the start angle of the arc")
        
def arc_start_angle_entry_event_cb(doc, text, tool):
    if len(text):
        _angle = make_angle(text)
        tool.setStartAngle(_angle)
        tool.setHandler("mouse_move", arc_angle_mouse_move_cb)
        tool.setHandler("left_button_press", arc_center_third_left_button_press_cb)
        tool.setHandler("entry_event", arc_end_angle_entry_event_cb)
        doc.setPrompt("Click in the drawing area or enter the end angle of the arc")
        
def arc_end_angle_entry_event_cb(doc, text, tool):
    if len(text):
        _angle = make_angle(text)
        tool.setEndAngle(_angle)
        create_entity(doc, tool)

        

#
# Chamfers
#
def chamfer_mode_init(doc, tool):
    doc.setPrompt("Click on the points where you want a chamfer")
    tool.setHandler("initialize", chamfer_mode_init)
    tool.setHandler("left_button_press", chamfer_left_button_press_cb)

def chamfer_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    _viewLoc = _da.convertPoint_fromView_(_loc, None)
    _tol = _da.pointSize().width
    _image = doc.getImage()
    _pt, _new_pt = _image.findPoint(_viewLoc.x, _viewLoc.y, _tol)
    if _pt is not None and not _new_pt:
        _active_layer = _image.getActiveLayer()
        _urefs = _pt.getUsers()
        if len(_urefs) != 2:
            return
        _s1 = _s2 = None        
        for _uref in _urefs:
            _user = _uref()
            if _user is None:
                return
            if not isinstance(_user, Generic.segment.Segment):
                return
            if _s1 is None:
                _s1 = _user
            else:
                _s2 = _user
        _len = _image.getOption('CHAMFER_LENGTH')
        Globals.s = _image.getOption('LINE_STYLE')
        Globals.l = _image.getOption('LINE_TYPE')
        Globals.c = _image.getOption('LINE_COLOR')
        Globals.t = _image.getOption('LINE_THICKNESS')
        #
        # the following is a work-around that needs to
        # eventually be replaced ...
        #
        _p1, _p2 = _s1.getEndpoints()
        _x1, _y1 = _p1.getCoords()
        _x2, _y2 = _p2.getCoords()
        _slen = _s1.length()
        _factor = 2.0
        _r = (_slen - (_len/_factor))/_slen
        _xn = _x1 + _r * (_x2 - _x1)
        _yn = _y1 + _r * (_y2 - _y1)
        _pc = _active_layer.find('point', _xn, _yn)
        while _pc is not None:
            _factor = _factor + 1.0
            if _factor > 1000:
                break
            _r = (_slen - (_len/_factor))/_slen
            _xn = _x1 + _r * (_x2 - _x1)
            _yn = _y1 + _r * (_y2 - _y1)
            _pc = _active_layer.find('point', _xn, _yn)
        if _pc is None:
            _pc = Generic.point.Point(_xn, _yn)
            _image.addObject(_pc)
            if _pt is _p1:
                _s1.setP1(_pc)
            else:
                _s1.setP2(_pc)
            _ptx, _pty = _pt.getCoords() 
            _pc.setx(_ptx)
            _pc.sety(_pty)
            _chamfer = Generic.segjoint.Chamfer(_s1, _s2, _len, Globals.s, Globals.l, Globals.c, Globals.t)
            _image.addObject(_chamfer)
            # force a redraw of the segments, too.
#            _msgr = doc.getMessenger()
#            _msgr.added_object_event(_image,_active_layer,_s1)
#            _msgr.added_object_event(_image,_active_layer,_s2)
            # we have to call another function to generate
            # an event to flush out the notification system
            # and draw what we just added.  This one works
            # ok, I guess.
            create_entity(doc, tool)
            return   
    
# 
# Fillets
#
def fillet_mode_init(doc, tool):
    doc.setPrompt("Click on the points where you want a fillet")
    tool.setHandler("initialize", fillet_mode_init)
    tool.setHandler("left_button_press", fillet_left_button_press_cb)
    
def fillet_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    _viewLoc = _da.convertPoint_fromView_(_loc, None)
    _tol = _da.pointSize().width
    _image = doc.getImage()
    _pt, _new_pt = _image.findPoint(_viewLoc.x, _viewLoc.y, _tol)
    if _pt is not None and not _new_pt:
        _active_layer = _image.getActiveLayer()
        _urefs = _pt.getUsers()
        if len(_urefs) != 2:
            return
        _s1 = _s2 = None        
        for _uref in _urefs:
            _user = _uref()
            if _user is None:
                return
            if not isinstance(_user, Generic.segment.Segment):
                return
            if _s1 is None:
                _s1 = _user
            else:
                _s2 = _user
        _rad = _image.getOption('FILLET_RADIUS')
        Globals.s = _image.getOption('LINE_STYLE')
        Globals.l = _image.getOption('LINE_TYPE')
        Globals.c = _image.getOption('LINE_COLOR')
        Globals.t = _image.getOption('LINE_THICKNESS')
        #
        # the following is a work-around that needs to
        # eventually be replaced ...
        #
        _p1, _p2 = _s1.getEndpoints()
        _x1, _y1 = _p1.getCoords()
        _x2, _y2 = _p2.getCoords()
        _slen = _s1.length()
        _factor = 2.0
        _r = 1.0 - (_slen/_factor)
        _xn = _x1 + _r * (_x2 - _x1)
        _yn = _y1 + _r * (_y2 - _y1)
        _pc = _active_layer.find('point', _xn, _yn)
        while _pc is not None:
            _factor = _factor + 1.0
            if _factor > 1000:
                break
            _r = 1.0 - (_slen/_factor)
            _xn = _x1 + _r * (_x2 - _x1)
            _yn = _y1 + _r * (_y2 - _y1)
            _pc = _active_layer.find('point', _xn, _yn)
        if _pc is None:
            _pc = Generic.point.Point(_xn, _yn)
            _image.addObject(_pc)
            if _pt is _p1:
                _s1.setP1(_pc)
            else:
                _s1.setP2(_pc)
            _ptx, _pty = _pt.getCoords()
            _pc.setx(_ptx)
            _pc.sety(_pty)
            _fillet = Generic.segjoint.Fillet(_s1, _s2, _rad, Globals.s, Globals.l, Globals.c, Globals.t)
            _image.addObject(_fillet)
            # we have to call another function to generate
            # an event to flush out the notification system
            # and draw what we just added.  This one works
            # ok, I guess.
            create_entity(doc, tool)
            return   
    
   
# 
# Leaders
#
def leader_mode_init(doc, tool):
    doc.setPrompt("Click in the drawing area to enter the first point")
    tool.setHandler("initialize", leader_mode_init)
    tool.setHandler("left_button_press", leader_first_left_button_press_cb)
    _image = doc.getImage()
    Globals.s = _image.getOption('LINE_STYLE')
    Globals.l = _image.getOption('LINE_TYPE')
    Globals.c = _image.getOption('LINE_COLOR')
    Globals.t = _image.getOption('LINE_THICKNESS')
    
def leader_first_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    _viewLoc = _da.convertPoint_fromView_(_loc, None)
    _da.setTempPoint(_viewLoc)
    tool.setFirstPoint(_viewLoc.x, _viewLoc.y)
    tool.setHandler("mouse_move", leader_mouse_move_cb)
    tool.setHandler("left_button_press", leader_second_left_button_press_cb)
    doc.setPrompt("Click in the drawing area to place the second point")
    
def leader_second_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    _viewLoc = _da.convertPoint_fromView_(_loc, None)
    _da.setTempPoint(_viewLoc)
    tool.setMidPoint(_viewLoc.x, _viewLoc.y)
    tool.setHandler("mouse_move", leader_mouse_move_cb)
    tool.setHandler("left_button_press", leader_third_left_button_press_cb)
    doc.setPrompt("Click in the drawing area to place the final point")
    
def leader_third_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    _viewLoc = _da.convertPoint_fromView_(_loc, None)
    tool.setFinalPoint(_viewLoc.x, _viewLoc.y)
    create_entity(doc, tool)

def leader_mouse_move_cb(doc, np, tool):
    _x1, _y1 = tool.getFirstPoint()
    _da = doc.getDA()
    _p2 = tool.getMidPoint()
    _flag = True
    if _p2 is not None:
        _s1 = Generic.segment.Segment((_x1, _y1), _p2, Globals.s, Globals.l, Globals.c, Globals.t)
        _da.setTempObject(_s1, _flag)
        _flag = False
        (_x1, _y1) = _p2
    (_x2, _y2) = np
    _s1 = Generic.segment.Segment((_x1, _y1), (_x2, _y2), Globals.s, Globals.l, Globals.c, Globals.t)
    _da.setTempObject(_s1, _flag)
    

#
# Polyline
#
def polyline_mode_init(doc, tool):
    doc.setPrompt("Click in the drawing area or enter the first point as 'x, y'")
    tool.setHandler("initialize", polyline_mode_init)
    tool.setHandler("left_button_press", polyline_left_button_press_cb)
    tool.setHandler("entry_event", polyline_entry_event_cb)
    _image = doc.getImage()
    Globals.s = _image.getOption('LINE_STYLE')
    Globals.l = _image.getOption('LINE_TYPE')
    Globals.c = _image.getOption('LINE_COLOR')
    Globals.t = _image.getOption('LINE_THICKNESS')
    
def polyline_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    _viewLoc = _da.convertPoint_fromView_(_loc, None)
    _da.setTempPoint(_viewLoc)
    tool.storePoint(_viewLoc.x, _viewLoc.y)
    if ((event.modifierFlags() & NSShiftKeyMask) != 0):
        create_entity(doc, tool)
    else:
        tool.setHandler("mouse_move", polyline_mouse_move_cb)
        doc.setPrompt("Click or enter as 'x, y' the next point. Shift-click or type 'last' to finish")

def polyline_mouse_move_cb(doc, np, tool):
    _pts = tool.getPoints()
    (_x1, _y1) = _pts[0]
    _count = len(_pts)
    _da = doc.getDA()
    _flag = True
    for _i in range(1, _count):
        (_x2, _y2) = _pts[_i]
        _s1 = Generic.segment.Segment((_x1, _y1), (_x2, _y2), Globals.s, Globals.l, Globals.c, Globals.t)
        _da.setTempObject(_s1, _flag)
        _flag = False
        _x1 = _x2
        _y1 = _y2
    (_x2, _y2) = np
    _s1 = Generic.segment.Segment((_x1, _y1), (_x2, _y2), Globals.s, Globals.l, Globals.c, Globals.t)
    _da.setTempObject(_s1, _flag)
        
def polyline_entry_event_cb(doc, text, tool):
    if len(text):
        if text == 'done':
            create_entity(doc, tool)
        else:
            _x, _y = make_tuple(text)
            tool.storePoint(_x, _y)
            tool.setHandler("mouse_move", polyline_mouse_move_cb)
            doc.setPrompt("Click or enter as 'x, y' the next point. Shift-click or type 'done' to finish")
            doc.getDA().setTempPoint(NSMakePoint(_x, _y))
        


#
# Polygon & External Polygon
#
def polygon_mode_init(doc, tool):
    doc.setPrompt("Click in the drawing area or enter a point as 'x, y'")
    tool.setHandler("initialize", polygon_mode_init)
    tool.setHandler("left_button_press", polygon_first_left_button_press_cb)
    tool.setHandler("entry_event", polygon_point_entry_event_cb)
    _image = doc.getImage()
    Globals.s = _image.getOption('LINE_STYLE')
    Globals.l = _image.getOption('LINE_TYPE')
    Globals.c = _image.getOption('LINE_COLOR')
    Globals.t = _image.getOption('LINE_THICKNESS')

def polygon_first_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    _viewLoc = _da.convertPoint_fromView_(_loc, None)
    tool.setCenter(_viewLoc.x, _viewLoc.y)
    tool.setHandler("left_button_press", polygon_second_left_button_press_cb)
    tool.setHandler("mouse_move", polygon_mouse_move_cb)
    tool.setHandler("entry_event", polygon_radius_entry_event_cb)
    doc.setPrompt("Click in the drawing area or enter the apparent radius")
    _da.setTempPoint(_viewLoc)
    
def polygon_second_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    (_x, _y) = _da.convertPoint_fromView_(_loc, None)
    (_cx, _cy) = tool.getCenter()
    if (_cx != _x) or (_cy != _y):
        tool.setLocation(_x, _y)
        create_entity(doc, tool)
   
def polygon_mouse_move_cb(doc, np, tool):
    _segs = []
    (_x, _y) = np
    tool.setLocation(_x, _y)
    _count = tool.getSideCount()
    _flag = True
    _da = doc.getDA()
    (_x0, _y0) = (_x1, _y1) = tool.getCoord(0)
    for _i in range(1, _count):
        (_x2, _y2) = tool.getCoord(_i)
        _s1 = Generic.segment.Segment((_x1, _y1), (_x2, _y2), Globals.s, Globals.l, Globals.c, Globals.t)
        _da.setTempObject(_s1, _flag)
        _x1 = _x2
        _y1 = _y2
        _flag = False
    _s1 = Generic.segment.Segment((_x0, _y0), (_x1, _y1), Globals.s, Globals.l, Globals.c, Globals.t)
    _da.setTempObject(_s1, False)

def polygon_point_entry_event_cb(doc, text, tool):
    if len(text):
        _x, _y = make_tuple(text)
        tool.setCenter(_x, _y)
        tool.setHandler("left_button_press", polygon_second_left_button_press_cb)
        tool.setHandler("mouse_move", polygon_mouse_move_cb)
        tool.delHandler("entry_event")
        doc.setPrompt("Click in the drawing area to complete the polygon")
        doc.getDA().setTempPoint(NSMakePoint(_x, _y))
    
def polygon_radius_entry_event_cb(doc, text, tool):
    if len(text):
        _r = make_radius(text)
        (_cx, _cy) = tool.getCenter()
        tool.setLocation(_cx+_r, _cy)
        create_entity(doc, tool)




#
# Copyright (c) 2004 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# A base class for objects that send messages
#

# import weakref
import types

class Messenger(object):
    messages = {
        'connected' : True,
        'disconnected' : True
        }
    def __init__(self):
        self.__connections = None
        self.__ignore = None
        self.__muted = False

    def connect(self, message, obj, method):
        if not self.sendsMessage(message):
            raise ValueError, "Unknown message : %s" % str(message)
        if not isinstance(method, types.MethodType):
            raise TypeError, "Invalid class method: %s" % str(method)
        if not isinstance(obj, method.im_class):
            raise ValueError, "Unknown method for object: %s" % str(method)
        if self.__connections is None:
            self.__connections = {}
        _tups = self.__connections.setdefault(message, [])
        _seen = False
        for _tup in _tups[:]:
            _obj = _tup[0]
            if _obj is obj:
                _seen = True
                break
        if not _seen:
            _tups.append((obj, method))
            self.sendMessage('connected', obj, method)

    def disconnect(self, obj, message=None):
        if self.__connections is not None:
            if message is None:
                _connections = self.__connections.keys()
                for _m in _connections:
                    _tups = self.__connections[_m]
                    for _tup in _tups[:]:
                        _obj = _tup[0]
                        if _obj is obj:
                            _tups.remove(_tup)
                            self.sendMessage('disconnected', obj)
                    if len(_tups) == 0:
                        del self.__connections[_m]
            else:
                if message in self.__connections:
                    _tups = self.__connections[message]
                    for _tup in _tups[:]:
                        _obj = _tup[0]
                        if _obj is obj:
                            _tups.remove(_tup)
                            self.sendMessage('disconnected', obj)
                            break
                    if len(_tups) == 0:
                        del self.__connections[message]
            if len(self.__connections) == 0:
                self.__connections = None

    def sendsMessage(self, m):
        return m in Messenger.messages

    def sendMessage(self, message, *args):
        if not isinstance(message, str):
            raise TypeError, "Message name not a string: " + `message`
        if not self.__muted and self.__connections is not None:
            if message in self.__connections:
                _tups = self.__connections[message][:] # make a copy
                for _obj, _method in _tups:
                    if isinstance(_obj, Messenger) and _obj.ignores(message):
                        continue
                    _method(_obj, self, *args)
    def mute(self):
        self.__muted = True

    def unmute(self):
        self.__muted = False

    def isMuted(self):
        return self.__muted

    def ignore(self, message):
        if not isinstance(message, str):
            raise TypeError, "Invalid message name: " + `message`
        if self.__ignore is None:
            self.__ignore = {}
        if message not in self.__ignore:
            self.__ignore[message] = True

    def receive(self, message):
        if not isinstance(message, str):
            raise TypeError, "Invalid message name: " + `message`
        if self.__ignore is not None:
            if message in self.__ignore:
                del self.__ignore[message]
            if len(self.__ignore) == 0:
                self.__ignore = None

    def ignores(self, message):
        return self.__ignore is not None and message in self.__ignore
        
    def receives(self, message):
        return self.__ignore is None or message not in self.__ignore



#
# Copyright (c) 2004 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# quadtree storage for dimension objects
#

import math

import dimension
import quadtree
import layer
import point
import circle
import arc
import tolerance

#
# Dimension storage
#

class DimQuadtree(quadtree.Quadtree):
    def __init__(self):
        quadtree.Quadtree.__init__(self)

    def addObject(self, obj):
        if not isinstance(obj, dimension.Dimension):
            raise TypeError, "Invalid Dimension object: " + `obj`
        if obj in self:
            return
        _dxmin, _dymin, _dxmax, _dymax = obj.getBounds()
        _node = self.getTreeRoot()
        _bounds = _node.getBoundary()
        _xmin = _ymin = _xmax = _ymax = None
        _resize = False
        if _bounds is None: # first node in tree
            _resize = True
            _xmin = _dxmin - 1.0
            _ymin = _dymin - 1.0
            _xmax = _dxmax + 1.0
            _ymax = _dymax + 1.0
        else:
            _xmin, _ymin, _xmax, _ymax = _bounds
            if _dxmin < _xmin:
                _xmin = _dxmin - 1.0
                _resize = True
            if _dxmax > _xmax:
                _xmax = _dxmax + 1.0
                _resize = True
            if _dymin < _ymin:
                _ymin = _dymin - 1.0
                _resize = True
            if _dymax > _ymax:
                _ymax = _dymax + 1.0
                _resize = True
        if _resize:
            self.resize(_xmin, _ymin, _xmax, _ymax)
            _node = self.getTreeRoot()
        _nodes = [_node]
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_dxmin > _xmax) or
                (_dymin > _ymax) or
                (_dxmax < _xmin) or
                (_dymax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _dxmax < _xmid: # dim on left side
                    _ne = _se = False
                if _dxmin > _xmid: # dim on right side
                    _nw = _sw = False
                if _dymax < _ymid: # dim below
                    _nw = _ne = False
                if _dymin > _ymid: # dim above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                if obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                    _node.addObject(obj)
        quadtree.Quadtree.addObject(self, obj)
        obj.connect('moved', self, DimQuadtree.moveDim)

    def delObject(self, obj):
        if obj not in self:
            return
        _dxmin, _dymin, _dxmax, _dymax = obj.getBounds()
        _xmin = _ymin = _xmax = _ymax = None
        _pdict = {}
        _nodes = [self.getTreeRoot()]
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_dxmin > _xmax) or
                (_dxmax < _xmin) or
                (_dymin > _ymax) or
                (_dymax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _dxmax < _xmid: # dim on left side
                    _ne = _se = False
                if _dxmin > _xmid: # dim on right side
                    _nw = _sw = False
                if _dymax < _ymid: # dim below
                    _nw = _ne = False
                if _dymin > _ymid: # dim above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                _node.delObject(obj) # dim may not be in the node ...
                _parent = _node.getParent()
                if _parent is not None:
                    _pid = id(_parent)
                    if _pid not in _pdict:
                        _pdict[_pid] = _parent
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        for _parent in _pdict.values():
            self.purgeSubnodes(_parent)

    def find(self, *args):
        _alen = len(args)
        if _alen < 4:
            raise ValueError, "Invalid argument count: %d" % _alen
        _l1 = args[0]
        if not isinstance(_l1, layer.Layer):
            raise TypeError, "Invalid layer: " + `_l1`
        _p1 = args[1]
        if not isinstance(_p1, point.Point):
            raise TypeError, "Invalid point: " + `_p1`
        _l2 = args[2]
        if not isinstance(_l2, layer.Layer):
            raise TypeError, "Invalid layer: " + `_l2`
        _p2 = args[3]
        if not isinstance(_p2, point.Point):
            raise TypeError, "Invalid point: " + `_p2`
        if not len(self):
            return None
        for _dim in self.getObjects():
            _dl1, _dl2 = _dim.getDimLayers()
            _dp1, _dp2 = _dim.getDimPoints()
            if ((_l1 is _dl1) and
                (_p1 is _dp1) and
                (_l2 is _dl2) and
                (_p2 is _dp2)):
                return _dim
            if ((_l1 is _dl2) and
                (_p1 is _dp2) and
                (_l2 is _dl1) and
                (_p2 is _dp1)):
                return _dim
        return None

    def moveDim(self, obj, *args):
        if obj not in self:
            raise ValueError, "Dimension not stored in Quadtree: " + `obj`
        _alen = len(args)
        if _alen < 4:
            raise ValueError, "Invalid argument count: %d" % _alen
        _dxmin = args[0]
        if not isinstance(_dxmin, float):
            _dxmin = float(args[0])
        _dymin = args[1]
        if not isinstance(_dymin, float):
            _dymin = float(args[1])
        _dxmax = args[2]
        if not isinstance(_dxmax, float):
            _dxmax = float(args[2])
        _dymax = args[3]
        if not isinstance(_dymax, float):
            _dymax = float(args[3])
        _nodes = [self.getTreeRoot()]
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_dxmin > _xmax) or
                (_dxmax < _xmin) or
                (_dymin > _ymax) or
                (_dymax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _dxmax < _xmid: # dim on left side
                    _ne = _se = False
                if _dxmin > _xmid: # dim on right side
                    _nw = _sw = False
                if _dymax < _ymid: # dim below
                    _nw = _ne = False
                if _dymin > _ymid: # dim above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                _node.delObject(obj) # dim may not be in node ...
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        self.addObject(obj)

    def getClosest(self, x, y, tol=tolerance.TOL):
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _dim = _tsep = None
        _bailout = False
        _ddict = {}
        _nodes = [self.getTreeRoot()]
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_x < (_xmin - _t)) or
                (_x > (_xmax + _t)) or
                (_y < (_ymin - _t)) or
                (_y > (_ymax + _t))):
                continue
            if _node.hasSubnodes():
                _nodes.extend(_node.getSubnodes())
            else:
                for _d in _node.getObjects():
                    _did = id(_d)
                    if _did not in _ddict:
                        _pt = _d.mapCoords(_x, _y, _t)
                        if _pt is not None:
                            _px = _py = _pt
                            _sep = math.hypot((_px - _x), (_py - _y))
                            if _tsep is None:
                                _tsep = _sep
                                _dim = _d
                            else:
                                if _sep < _tsep:
                                    _tsep = _sep
                                    _dim = _d
                            if _dim is not None and _sep < 1e-10:
                                _bailout = True
                                break
            if _bailout:
                break
        return _dim

    def getInRegion(self, xmin, ymin, xmax, ymax):
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        _dims = []
        if not len(self):
            return _dims
        _nodes = [self.getTreeRoot()]
        _ddict = {}
        while len(_nodes):
            _node = _nodes.pop()
            if _node.hasSubnodes():
                for _subnode in _node.getSubnodes():
                    _sxmin, _symin, _sxmax, _symax = _subnode.getBoundary()
                    if ((_sxmin > _xmax) or
                        (_symin > _ymax) or
                        (_sxmax < _xmin) or
                        (_symax < _ymin)):
                        continue
                    _nodes.append(_subnode)
            else:
                for _dim in _node.getObjects():
                    _dimid = id(_dim)
                    if _dimid not in _ddict:
                        if _dim.inRegion(_xmin, _ymin, _xmax, _ymax):
                            _dims.append(_dim)
                        _ddict[_dimid] = True
        return _dims

#
# LinearDimension Quadtree
#

class LDimQuadtree(DimQuadtree):
    def __init__(self):
        DimQuadtree.__init__(self)

    def addObject(self, obj):
        if not isinstance(obj, dimension.LinearDimension):
            raise TypeError, "Invalid LinearDimension object: " + `obj`
        DimQuadtree.addObject(self, obj)

#
# HorizontalDimension Quadtree
#

class HDimQuadtree(DimQuadtree):
    def __init__(self):
        DimQuadtree.__init__(self)

    def addObject(self, obj):
        if not isinstance(obj, dimension.HorizontalDimension):
            raise TypeError, "Invalid HorizontalDimension object: " + `obj`
        DimQuadtree.addObject(self, obj)

#
# VerticalDimension Quadtree
#

class VDimQuadtree(DimQuadtree):
    def __init__(self):
        DimQuadtree.__init__(self)

    def addObject(self, obj):
        if not isinstance(obj, dimension.VerticalDimension):
            raise TypeError, "Invalid VerticalDimension object: " + `obj`
        DimQuadtree.addObject(self, obj)

#
# RadialDimension Quadtree
#

class RDimQuadtree(DimQuadtree):
    def __init__(self):
        DimQuadtree.__init__(self)

    def addObject(self, obj):
        if not isinstance(obj, dimension.RadialDimension):
            raise TypeError, "Invalid RadalDimension object: " + `obj`
        DimQuadtree.addObject(self, obj)

    def find(self, *args):
        _alen = len(args)
        if _alen < 2:
            raise ValueError, "Invalid argument count: %d" % _alen
        _l1 = args[0]
        if not isinstance(_l1, layer.Layer):
            raise TypeError, "Invalid layer: " + `_l1`
        _c1 = args[1]
        if not isinstance(_c1, (circle.Circle, arc.Arc)):
            raise TypeError, "Invalid circle/arc: " + `_c1`
        if not len(self):
            return None
        for _rdim in self.getObjects():
            _rl = _rdim.getDimLayer()
            _rc = _rdim.getDimCircle()
            if _rl is _l1 and _c1 is _rc:
                return _rdim
        return None

#
# AngularDimension Quadtree
#

class ADimQuadtree(DimQuadtree):
    def __init__(self):
        DimQuadtree.__init__(self)

    def addObject(self, obj):
        if not isinstance(obj, dimension.AngularDimension):
            raise TypeError, "Invalid AngularDimension object: " + `obj`
        DimQuadtree.addObject(self, obj)

    def find(self, *args):
        _alen = len(args)
        if _alen < 6:
            raise ValueError, "Invalid argument count: %d" % _alen
        _vl = args[0]
        if not isinstance(_vl, layer.Layer):
            raise TypeError, "Invalid layer: " + `_vl`
        _vp = args[1]
        if not isinstance(_vp, point.Point):
            raise TypeError, "Invalid point: " + `_vp`
        _l1 = args[2]
        if not isinstance(_l1, layer.Layer):
            raise TypeError, "Invalid layer: " + `_l1`
        _p1 = args[3]
        if not isinstance(_p1, point.Point):
            raise TypeError, "Invalid point: " + `_p1`
        _l2 = args[4]
        if not isinstance(_l2, layer.Layer):
            raise TypeError, "Invalid layer: " + `_l2`
        _p2 = args[5]
        if not isinstance(_p2, point.Point):
            raise TypeError, "Invalid point: " + `_p2`
        if not len(self):
            return None
        for _adim in self.getObjects():
            _avl, _al1, _al2 = _adim.getDimLayers()
            _avp, _ap1, _ap2 = _adim.getDimPoints()
            if ((_vl is _avl) and
                (_vp is _avp) and
                (_l1 is _al1) and
                (_p1 is _ap1) and
                (_l2 is _al2) and
                (_p2 is _ap2)):
                return _adim
            if ((_vl is _avl) and
                (_vp is _avp) and
                (_l1 is _al2) and
                (_p1 is _ap2) and
                (_l2 is _al1) and
                (_p2 is _ap1)):
                return _adim
        return None



#
# Copyright (c) 2003 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# This code handles interpreting entries in the gtkimage.entry box
# and calling the apprpriate internal command
#
# Author: David Broadwell ( dbroadwell@mindspring.com, 05/26/2003 )
#


''' 05/29/03 R2c Complete redesign, no longer does the internal name
    need to be a function, dropped prompt.py filesize from 9.8 to
    5.3Kb! cleaned up the feywords file, no appreciable difference
    Used a dictionary instead of discreet functions for every entry
    and for R3, menus. '''


def lookup(text):
    """Interface to promptdef dictionary, returns code by keyword definition

lookup(text)    
    """
    assert text in promptdefs, "No command for %s" % str(text)
    return promptdefs[text]

promptdefs = {
    'acline' : "NSApp.sendAction_to_from_(\"drawACLine:\", None, self)",
    'arcc' : "NSApp.sendAction_to_from_(\"drawArc:\", None, self)",
    'ccircen' : "NSApp.sendAction_to_from_(\"drawCCircleCenter:\", None, self)",
    'ccir2p' : "NSApp.sendAction_to_from_(\"drawCCircle2Point:\", None, self)",
    'chamfer' : "NSApp.sendAction_to_from_(\"drawChamfer:\", None, self)",
    'close' : "NSApp.sendAction_to_from_(\"performClose:\", None, self)",
    'circen' : "NSApp.sendAction_to_from_(\"drawCircleCentered:\", None, self)",
    'cir2p' : "NSApp.sendAction_to_from_(\"drawCircle2Point:\", None, self)",
    'cl' : "NSApp.sendAction_to_from_(\"draw2PointCLine:\", None, self)",
    'cut' : "NSApp.sendAction_to_from_(\"cut:\", None, self)",
    'copy' : "NSApp.sendAction_to_from_(\"copy:\", None, self)",
    'delete' : "NSApp.sendAction_to_from_(\"delete:\", None, self)",
    'dimpref' : "NSApp.sendAction_to_from_(\"dimension_prefs_cb(None)",
    'fillet' : "NSApp.sendAction_to_from_(\"drawFillet:\", None, self)",
    'hcline' : "NSApp.sendAction_to_from_(\"drawHCLine:\", None, self)",
    'leader' : "NSApp.sendAction_to_from_(\"drawLeader:\", None, self)",
    'mirror' : "NSApp.sendAction_to_from_(\"modifyMirror:\", None, self)",
    'moveh' : "NSApp.sendAction_to_from_(\"modifyMoveHorizontal:\", None, self)",
    'movev' : "NSApp.sendAction_to_from_(\"modifyMoveVertical:\", None, self)",
    'move' : "NSApp.sendAction_to_from_(\"modifyMoveFree:\", None, self)",
    'new' : "NSApp.sendAction_to_from_(\"newDocument:\", None, self)",
    'opend' : "NSApp.sendAction_to_from_(\"openDocument:\", None, self)",
    'paste' : "NSApp.sendAction_to_from_(\"paste:\", None, self)",
    'pcline' : "NSApp.sendAction_to_from_(\"drawPerpendicularCLine:\", None, self)",
    'point' : "NSApp.sendAction_to_from_(\"drawPoint:\", None, self)",
    'polyline' : "NSApp.sendAction_to_from_(\"drawPolyline:\", None, self)",
    'polygon' : "NSApp.sendAction_to_from_(\"drawPolygon:\", None, self)", 
    'polygonext' : "NSApp.sendAction_to_from_(\"drawPolygonExternal:\", None, self)", 
    'pref' : "NSApp.sendAction_to_from_(\"prefs_cb(None)",
    'rect' : "NSApp.sendAction_to_from_(\"drawRect:\", None, self)",
    'redraw' : "self.document().getDA().setNeedsDisplay_(True)",
    'redo ' : "NSApp.delegate().redo_(None)",
    'refresh' : "self.document().getDA().setNeedsDisplay_(True)",
    'saa' : "NSApp.sendAction_to_from_(\"select_all_arcs_cb(None)",
    'saacl' : "NSApp.sendAction_to_from_(\"select_all_aclines_cb(None)",
    'sac' : "NSApp.sendAction_to_from_(\"select_all_circles_cb(None)",
    'sacc' : "NSApp.sendAction_to_from_(\"select_all_ccircles_cb(None)",
    'sacl' : "NSApp.sendAction_to_from_(\"select_all_clines_cb(None)",
    'sahcl' : "NSApp.sendAction_to_from_(\"select_all_hclines_cb(None)",
    'sap' : "NSApp.sendAction_to_from_(\"select_all_points_cb(None)",
    'sas' : "NSApp.sendAction_to_from_(\"select_all_segments_cb(None)",
    'savcl' : "NSApp.sendAction_to_from_(\"select_all_vclines_cb(None)",
    'saveas' : "NSApp.sendAction_to_from_(\"saveDocumentAs:\", None, self)",
    'saves' : "NSApp.sendAction_to_from_(\"saveDocument:\", None, self)",
    'savel' : "NSApp.sendAction_to_from_(\"saveLayerAs:\", None, self)",
    'segment' : "NSApp.sendAction_to_from_(\"drawSegment:\", None, self)",
    'split' : "NSApp.sendAction_to_from_(\"modifySplit:\", None, self)",
    'strh' : "NSApp.sendAction_to_from_(\"modifyStretchHorizontal:\", None, self)",
    'strv' : "NSApp.sendAction_to_from_(\"modifyStretchVertical:\", None, self)",
    'str' : "NSApp.sendAction_to_from_(\"modifyStretchFree:\", None, self)",
    'tcline' : "NSApp.sendAction_to_from_(\"drawTanCLine:\", None, self)",
    'text' : "NSApp.sendAction_to_from_(\"drawText:\", None, self)",
    'transfer' : "NSApp.sendAction_to_from_(\"modifyTransfer:\", None, self)",
    'undo' : "NSApp.delegate().undo_(None)",
    'vcline' : "NSApp.sendAction_to_from_(\"drawVCLine:\", None, self)",
    'quit' : "NSApp.sendAction_to_from_(\"terminate:\", None, self)",
    'zoomd' : "NSApp.sendAction_to_from_(\"windowZoom:\", None, self)",
    'zoomi' : "NSApp.sendAction_to_from_(\"windowZoomIn:\", None, self)",
    'zoomo' : "NSApp.sendAction_to_from_(\"windowZoomOut:\", None, self)",
    'zoomf' : "NSApp.sendAction_to_from_(\"windowZoomFit:\", None, self)"
}



#
# Copyright (c) 2003 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# This code sets the global entries for the gtkimage.entry box
# commands text to 'defined' behavior for an application set
# thus simulating a mainstream CAD package at the users need
#
# Author: David Broadwell ( dbroadwell@mindspring.com, 05/26/2003 )
#

"""
Defaultglobals is loaded by default, use it's pattern to make
your own behave like the CAD you are familiar with.

Please note the example below;

exampledefinitions = {
    'foobar' : "quit()",
    'foo' : "segment()",
    'X' : "quit()",
    'O' : "opend()"
}
"""

defaultglobals = {
    'acline' : "acline",
    'acl' : "acline",
    'arc' : "arcc",
    'ccir2' : "ccir2p",
    'ccir' : "ccircen",
    'chamfer' : "chamfer",
    'cir2' : "cir2p",
    'cir' : "circen",
    'cl' : "cl",
    'close' : "close",
    'copy' : "copy",
    'cut' : "cut",
    'delete' : "delete",
    'del' : "delete",
    'dimpref' : "dimpref",
    'exit' : "quit",
    'fillet' : "fillet",
    'hcline' : "hcline",
    'hcl' : "hcline",
    'leader' : "leader",
    'lead' : "leader",
    'mirror' : "mirror",
    'mir' : "mirror",
    'move' : "move",
    'mv' : "move",
    'moveh' : "moveh",
    'movev' : "movev",
    'new' : "new",
    'opend' : "opend",
    'paste' : "paste",
    'pcline' : "pcline",
    'point' : "point",
    'polyline' : "polyline",
    'pline' : "polyline",
    'pref' : "pref",
    'quit' : "quit",
    'rectangle' : "rect",
    'rect' : "rect",
    'redraw' : "redraw",
    'refresh' : "refresh",
    'r' : "refresh",
    'saa' : "saa",
    'saacl' : "saacl",
    'sac' : "sac",
    'sacc' : "sacc",
    'sacl' : "sacl",
    'sahcl' : "sahcl",
    'sap' : "sap",
    'sas' : "sas",
    'savcl' : "savcl",
    'saveas' : "saveas",
    'savel' : "savel",
    'saves' : "saves",
    'sv' : "saves",
    'segment' : "segment",
    'seg' : "segment",
    'l' : "segment",
    'split' : "split",
    'str' : "str",
    'strh' : "strh",
    'strv' : "strv",
    'tcline' : "tcline",
    'tcl' : "tcline",
    'text' : "text",
    'transfer' : "transfer",
    'vcline' : "vcline",
    'vcl' : "vcline",
    'zoomd' : "zoomd",
    'z' : "zoomd",
    'zoomf' : "zoomf",
    'zf' : "zoomf",
    'zoomi' : "zoomi",
    'zi' : "zoomi",
    'zoomo' : "zoomo",
    'zo' : "zoomo"
}

acadglobals = defaultglobals

# me10globals = defaultglobals




#
# Copyright (c) 2002, 2003, 2004 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# classes for graphic objects
#

import warnings

import style
import color
import linetype
import baseobject
import logger

class old_GraphicObject(object):
    """A class representing an object that is drawn.

This class is meant to be a base class for things like
line segments, circles, etc. The class has one attribute
that cannot be None:

style: The Style object

There are three other attributes that override values in
the style attribute:

color: A Color object
thickness: A positive float value
linetype: A Linetype object

A GraphicObject object has the following methods:

{get/set}Style(): Get/Set the Style of a GraphicObject.
{get/set}Color(): Get/Set the Color of a GraphicObject.
{get/set}Thickness(): Get/Set the thickness of a GraphicObject.
{get/set}Linetype(): Get/Set the Linetype of a GraphicObject.
inRegion(): Returns whether or not the GraphicObject is within a bounded area.
    """

    __defstyle = style.Style(u'Default Style',
                             linetype.Linetype(u'Solid', None),
                             color.Color(0xffffff),
                             1.0)

    def __init__(self, st=None, lt=None, col=None, t=None):
        """Initialize a GraphicObject.

GraphicObject([st, lt, col, t])

Optional arguments:
st: A style object that overrides the class default
lt: A Linetype object that overrides the value in the Style
col: A Color object that overrides the value in the Style
t: A positive float that overrides the value in the Style
        """
        _st = st
        if _st is not None:
            if not isinstance(_st, style.Style):
                raise TypeError, "Invalid style: " + `_st`
        _col = col
        if _col is not None:
            if not isinstance(_col, color.Color):
                _col = color.Color(col)
        _lt = lt
        if _lt is not None:
            if not isinstance(_lt, linetype.Linetype):
                raise TypeError, "Invalid linetype: " + `_lt`
        _t = t
        if _t is not None:
            if not isinstance(_t, float):
                _t = float(t)
            if _t < 0.0:
                raise ValueError, "Invalid thickness: %g" % _t
        self.__style = _st
        self.__color = _col
        self.__linetype = lt
        self.__thickness = _t
        
        
    def getStyle(self):
        """Get the Style of the GraphicObject.

getStyle()

Returns the current Style of an object. The values in the
style may be overriden if any of the color, linetype, or
thickness attributes are not None.
        """
        _style = self.__style
        if _style is None:
            _style = old_GraphicObject.__defstyle
        return _style

    def setStyle(self, s):
        """Set the Style of the Object.

setStyle([s])

Setting the style of a GraphicObject unsets any overrides for
the object. Setting the Style to None restores the default
style.
        """
        _s = s
        if _s is not None:
            if not isinstance(_s, style.Style):
                raise TypeError, "Invalid style: " + `_s`
        self.__style = _s

    style = property(getStyle, setStyle, None, "Object Style.")

    def getColor(self):
        """Get the Color of the Object.

getColor()
        """
        _color = self.__color
        if _color is None:
            if self.__style is not None:
                _color = self.__style.getColor()
        if _color is None:
            _color = old_GraphicObject.__defstyle.getColor()
        return _color

    def setColor(self, c=None):
        """Set the color of the object.

setColor([c])

Setting the color overrides the value in the Style. Setting
the color to None or invoking this method without arguments
restores the color value defined in the Style.
        """
        _c = c
        if _c is not None:
            if not isinstance(_c, color.Color):
                _c = color.Color(c)
        self.__color = _c
        
    color = property(getColor, setColor, None, "Object color.")
    
    def getThickness(self):
        """Get the thickness of the GraphicObject.

getThickness()
        """
        _th = self.__thickness
        if _th is None:
            if self.__style is not None:
                _th = self.__style.getThickness()
        if _th is None:
            _th = old_GraphicObject.__defstyle.getThickness()
        return _th

    def setThickness(self, t=None):
        """Set the thickness of a object.

setThickness([t])

Setting the thickness overrides the value in the Style. Setting
the thickness to None, or invoking this method without an
argument, restores the thickness value defined in the Style.
        """
        _t = t
        if _t is not None:
            if not isinstance(_t, float):
                _t = float(t)
            if _t < 0.0:
                raise ValueError, "Invalid thickness: %g" % _t
        self.__thickness = _t

    thickness = property(getThickness, setThickness, None, "Object thickness.")
    
    def getLinetype(self):
        """Get the Linetype of the object.

getLinetype()
        """
        _lt = self.__linetype
        if _lt is None:
            if self.__style is not None:
                _lt = self.__style.getLinetype()
        if _lt is None:
            _lt = old_GraphicObject.__defstyle.getLinetype()
        return _lt

    def setLinetype(self, lt=None):
        """Set the Linetype of the GraphicObject.

setLinetype([lt])

Setting the Linetype overrides the value in the Sytle. Setting
the Linetype to None or invoking this method without arguments
restores the linetype value defined in the Style.
        """
        _lt = lt
        if _lt is not None:
            if not isinstance(_lt, linetype.Linetype):
                raise TypeError, "Invalid linetype: " + `_lt`
        self.__linetype = _lt

    linetype = property(getLinetype, setLinetype, None, "Object Linetype.")

    def inRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return whether or not an object falls within a boundary.

isRegion(xmin, ymin, xmax, ymax[, fully])

The first four arguments define the boundary. The optional
fifth argument "fully" indicates whether or not the object
must be completely contained within the region or just pass
through it.

This method should be overriden in classes derived from GraphicObjects.
        """
        return False

class GraphicObject(baseobject.Subpart):
    """A class representing an object that is drawn.

This class is meant to be a base class for things like
line segments, circles, etc. The GraphicObject class
has one attribute that cannot be None:

style: The Style object

There are three other attributes that override values in
the style attribute:

color: A Color object
thickness: A positive float value
linetype: A Linetype object

A GraphicObject object has the following methods:

{get/set}Style(): Get/Set the Style of a GraphicObject.
{get/set}Color(): Get/Set the Color of a GraphicObject.
{get/set}Thickness(): Get/Set the thickness of a GraphicObject.
{get/set}Linetype(): Get/Set the Linetype of a GraphicObject.
    """
    messages = {
    'attribute_changed' : True,
    }
    __defstyle = style.Style(u'Default Style',
                             linetype.Linetype(u'Solid', None),
                             color.Color(0xffffff),
                             1.0)

    def __init__(self, st=None, lt=None, col=None, t=None, **kw):
        """Initialize a GraphicObject.

GraphicObject([st, lt, col, t])

Optional arguments:
style: A style object that overrides the class default
linetype: A Linetype object that overrides the value in the Style
color: A Color object that overrides the value in the Style
thickness: A positive float that overrides the value in the Style
        """
        baseobject.Subpart.__init__(self, **kw)
        _st = st
        if _st is None:
            _st = GraphicObject.__defstyle
        else:
            if not isinstance(_st, style.Style):
                raise TypeError, "Invalid style: " + `_st`
        _col = _lt = _t = None
        if 'color' in kw:
            _col = kw['color']
        else:
            if 'col' in kw:
                warnings.warn("keyword 'col' is deprecated;  use 'color'",
                DeprecationWarning, stacklevel=2)
                _col = kw['col']
        if _col is not None:
            if not isinstance(_col, color.Color):
                _col = color.Color(_col)
            if _col == _st.getColor():
                _col = None
        _col = col
        if _col is not None:
            if not isinstance(_col, color.Color):
                _col = color.Color(col)
            if _col == _st.getColor():
                _col = None
        if 'linetype' in kw:
            _lt = kw['linetype']
        else:
            if 'lt' in kw:
                warnings.warn("keyword 'lt' is deprecated;  use 'linetype'",
                              DeprecationWarning, stacklevel=2)
                _lt = kw['lt']
        if _lt is not None:
            if not isinstance(_lt, linetype.Linetype):
                raise TypeError, "Invalid linetype: " + `_lt`
            if _lt == _st.getLinetype():
                _lt = None
        _lt = lt
        if _lt is not None:
            if not isinstance(_lt, linetype.Linetype):
                raise TypeError, "Invalid linetype: " + `_lt`
            if _lt == _st.getLinetype():
                _lt = None
        if 'thickness' in kw:
            _t = kw['thickness']
        else:
            if 'th' in kw:
                warnings.warn("keyword 'th' is deprecated;  use 'thickness'",
                               DeprecationWarning, stacklevel=2)
                _t = kw['th']
        if _t is not None:
            if not isinstance(_t, float):
                _t = float(_t)
            if _t < 0.0:
                raise ValueError, "Invalid thickness: %g" % _t
            if abs(_t - _st.getThickness()) < 1e-10:
                _t = None
        _t = t
        if _t is not None:
            if not isinstance(_t, float):
                _t = float(t)
            if _t < 0.0:
                raise ValueError, "Invalid thickness: %g" % _t
            if abs(_t - _st.getThickness()) < 1e-10:
                _t = None
        self.__style = _st
        self.__color = _col
        self.__linetype = _lt
        self.__thickness = _t
        
    def getStyle(self):
        """Get the Style of the GraphicObject.

getStyle()

Returns the current Style of an object. The values in the
style may be overriden if any of the color, linetype, or
thickness attributes are not None.
        """
        _style = self.__style
        if _style is None:
            _style = GraphicObject.__defstyle
        return _style

    def setStyle(self, s):
        """Set the Style of the Object.

setStyle([s])

Setting the style of a GraphicObject unsets any overrides for
the object. Setting the Style to None restores the default
style.
        """
        if self.isLocked():
            raise RuntimeError, "Style change not allowed - objected locked."
        _s = s
        if _s is not None:
            if not isinstance(_s, style.Style):
                raise TypeError, "Invalid style: " + `_s`
        _cs = self.getStyle()            
        if _cs != _s:
            self.__style = _s
            self.sendMessage('attribute_changed', 'style', _cs)
            self.modified()

    style = property(getStyle, setStyle, None, "Object Style.")

    def getColor(self):
        """Get the Color of the Object.

getColor()
        """
        _color = self.__color
        if _color is None:
            if self.__style is not None:
                _color = self.__style.getColor()
        if _color is None:
            _color = GraphicObject.__defstyle.getColor()
        return _color

    def setColor(self, c=None):
        """Set the color of the object.

setColor([c])

Setting the color overrides the value in the Style. Setting
the color to None or invoking this method without arguments
restores the color value defined in the Style.
        """
        if self.isLocked():
            raise RuntimeError, "Color change not allowed - object locked."
        _c = c
        if _c is not None:
            if not isinstance(_c, color.Color):
                _c = color.Color(c)
        _oc = self.getColor()                
        if _oc != _c:
            self.__color = _c
            self.sendMessage('attribute_changed', 'color', _oc)
            self.modified()
        
    color = property(getColor, setColor, None, "Object color.")
    
    def getThickness(self):
        """Get the thickness of the GraphicObject.

getThickness()
        """
        _th = self.__thickness
        if _th is None:
            if self.__style is not None:
                _th = self.__style.getThickness()
        if _th is None:
            _th = GraphicObject.__defstyle.getThickness()
        return _th

    def setThickness(self, t=None):
        """Set the thickness of a object.

setThickness([t])

Setting the thickness overrides the value in the Style. Setting
the thickness to None, or invoking this method without an
argument, restores the thickness value defined in the Style.
        """
        if self.isLocked():
            raise RuntimeError, "Thickness change not allowed - object locked."
        _t = t
        if _t is not None:
            if not isinstance(_t, float):
                _t = float(t)
            if _t < 0.0:
                raise ValueError, "Invalid thickness: %g" % _t
        _ot = self.getThickness()            
        if ((_t is None and _ot is not None) or
            (_t is not None and _ot is None) or
            (_t is not None and _ot is not None and abs(_ot - _t) > 1e-10)):
            self.__thickness = _t
            self.sendMessage('attribute_changed', 'thickness', _ot)
            self.modified()

    thickness = property(getThickness, setThickness, None, "Object thickness.")
    
    def getLinetype(self):
        """Get the Linetype of the object.

getLinetype()
        """
        _lt = self.__linetype
        if _lt is None:
            if self.__style is not None:
                _lt = self.__style.getLinetype()
        if _lt is None:
            _lt = GraphicObject.__defstyle.getLinetype()
        return _lt

    def setLinetype(self, lt=None):
        """Set the Linetype of the GraphicObject.

setLinetype([lt])

Setting the Linetype overrides the value in the Sytle. Setting
the Linetype to None or invoking this method without arguments
restores the linetype value defined in the Style.
        """
        if self.isLocked():
            raise RuntimeError, "Linetype change not allowed - object locked."
        _lt = lt
        if _lt is not None:
            if not isinstance(_lt, linetype.Linetype):
                raise TypeError, "Invalid linetype: " + `_lt`
        _ol = self.getLinetype()            
        if _ol != _lt:
            self.__linetype = _lt
            self.sendMessage('attribute_changed', 'linetype', _ol)
            self.modified()

    linetype = property(getLinetype, setLinetype, None, "Object Linetype.")

    def getGraphicValues(self, defstyle=None):
        if defstyle is not None and not isinstance(defstyle, style.Style):
            raise TypeError, "Invalid default style: " + `defstyle`
        _s = _c = _l = _t = None        
        _ss = self.__style        
        if defstyle is None or defstyle != _ss:
            _s = _ss.getStyleValues()
        _sc = self.__color
        if _sc is not None:
            _c = str(_sc)
        _sl = self.__linetype
        if _sl is not None:
            _l = (_sl.getName(), _sl.getList())
        _st = self.__thickness
        if _st is not None:
            _t = self.__thickness
        return _s, _l, _c, _t
        
    def sendsMessage(self, m):
        if m in GraphicObject.messages:
            return True
        return baseobject.Subpart.sendsMessage(self, m)

#
# GraphicObject history class
#

class GraphicObjectLog(logger.Logger):
    def __init__(self, obj):
        if not isinstance(obj, GraphicObject):
            raise TypeError, "Invalid GraphicObject: " + `obj`
        logger.Logger.__init__(self)
        self.__obj = obj
        obj.connect('attribute_changed', self, GraphicObjectLog.attrChanged)

    def detatch(self):
        self.__obj.disconnect(self)
        self.__obj = None

    def getObject(self):
        return self.__obj
    
    def attrChanged(self, obj, *args):
        _alen = len(args)
        if len(args) < 2:
            raise ValueError, "Invalid argument count: %d" % _alen
        _attr = args[0]
        if not isinstance(_attr, str):
            raise TypeError, "Unexpected attribute string: " + str(_attr)
        _v = args[1]
        if _attr == 'linetype':
            if not isinstance(_v, linetype.Linetype):
                raise TypeError, "Invalid linetype: " + `_v`
            _val = (_v.getName(), _v.getList())
        elif _attr == 'color':
            if not isinstance(_v, color.Color):
                raise TypeError, "Invalid color: " + `_v`
            _val = (_v.r, _v.g, _v.b)
        elif _attr == 'style':
            if not isinstance(_v, style.Style):
                raise TypeError, "Invalid style: " + `_v`
            _l = _v.getLinetype.getList()
            _c = str(_v.getColor())
            _t = _v.getThickness()
            _val = (_l, _c, _t)
        elif _attr == 'thickness':
            _val = _v
            if not isinstance(_val, float):
                _val = float(_v)
        else:
            raise ValueError, "Unexpected attribute: %s" % _attr
        self.saveUndoData('attr_changed', _attr, _val)

    def execute(self, action, *args):
        if action != 'undo' and action != 'redo':
            raise ValueError, "Unexpected action: %s" % action
        _alen = len(args)
        if len(args) == 0:
            raise ValueError, "No arguments to execute()"
        _obj = self.__obj
        _op = args[0]
        if _op == 'attr_changed':        
            if len(args) < 3:
                raise ValueError, "Invalid argument count: %d" % _alen
            _attr = args[1]
            _val = args[2]
            if _attr == 'linetype':
                _sv = (_obj.getName(), _obj.getList())
            elif _attr == 'color':
                _col = _obj.getColor()
                _sv = (_col.r, _col.g, _col.b)
            elif _attr == 'style':
                _l = _obj.getLinetype.getList()
                _c = str(_obj.getColor())
                _t = _obj.getThickness()
                _sv = (_l, _c, _t)
            elif _attr == 'thickness':
                _sv = _obj.getThickness()
            else:
                raise ValueError, "Unexpected attribute: %s" % _attr
            if action == 'undo':
                self.saveRedoData(_op, _attr, _sv)
            else:
                self.saveUndoData(_op, _attr, _sv)
            self.ignore('attribute_changed')
            try:
                if _attr == 'style':
                    pass # fixme
                elif _attr == 'linetype':
                    _name, _list = _val
                    _lt = linetype.get_linetype_by_dashes(_list)
                    if undo:
                        _obj.startUndo(True)
                        try:
                            _obj.setLinetype(_lt)
                        finally:
                            _obj.endUndo()
                    else:
                        _obj.setLinetype(_lt)
                elif _attr == 'color':
                    _r, _g, _b = _val
                    _col = color.get_color(_r, _g, _b)
                    if undo:
                        _obj.startUndo(True)
                        try:
                            _obj.setColor(_col)
                        finally:
                            _obj.endUndo()
                    else:
                        _obj.setColor(_col)
                elif _attr == 'thickness':
                    if undo:
                        _obj.startUndo(True)
                        try:
                            _obj.setThickness(_val)
                        finally:
                            _obj.endUndo()
                    else:
                        _obj.setThickness(_val)
                else:
                    raise ValueError, "Unexpected attribute: %s" % _attr
            finally:
                self.receive('attribute_changed')
        else:
            raise ValueError, "Unexpected operation: %s" % _op
